<?php
namespace HCM\Geral\Services;

use Doctrine\ORM\EntityManager;
use HCM\Util\{
    DateUtil,
    FileUploader,
    HCMParameters,
    Repositories,
    Messages,
    Util
};
use Teknisa\Model\Gestaopessoas\Entities\{GpeAnexos, GpeMigracao};
use HCM\Geral\Logic\ImportacaoXLS;

class MigracaoPlanilhaService
{
    const FILE_SERVER_BASE_PATH = "/migracao/planilha";
    const DEFAULT_MIGRACAO_IDSTATUS = "P";
    const DEFAULT_MODELO_FILENAME = "templates-excel-migracao.zip";

    const TIPO_MIGRACAO_CONFIG = [
        [
            "id" => 1,
            "order" => 1,
            "name" => "Estrutura",
            "fileServerPath" => "/1_Estrutura",
        ],
        [
            "id" => 2,
            "order" => 2,
            "name" => "Ocupação",
            "fileServerPath" => "/2_Ocupacao",
        ],
        [
            "id" => 3,
            "order" => 3,
            "name" => "Escala de Trabalho",
            "fileServerPath" => "/3_EscalaTrabalho",
        ],
        [
            "id" => 4,
            "order" => 4,
            "name" => "Vínculo",
            "fileServerPath" => "/4_Vinculo",
        ],
        [
            "id" => 5,
            "order" => 5,
            "name" => "Dependentes",
            "fileServerPath" => "/5_Dependentes",
        ],
        [
            "id" => 6,
            "order" => 6,
            "name" => "Alteração Salarial",
            "fileServerPath" => "/6_AlteracaoSalarial",
        ],
        [
            "id" => 7,
            "order" => 7,
            "name" => "Alteração de Escala",
            "fileServerPath" => "/7_AlteracaoEscala",
        ],
        [
            "id" => 8,
            "order" => 8,
            "name" => "Alteração de Ocupação",
            "fileServerPath" => "/8_AlteracaoOcupacao",
        ],
        [
            "id" => 9,
            "order" => 9,
            "name" => "Situação Funcional",
            "fileServerPath" => "/9_SituacaoFuncional",
        ],
        [
            "id" => 10,
            "order" => 10,
            "name" => "Férias",
            "fileServerPath" => "/10_Ferias",
        ],
        [
            "id" => 11,
            "order" => 11,
            "name" => "Movimentações de Estrutura",
            "fileServerPath" => "/11_MovimentacoesEstrutura",
        ],
        [
            "id" => 12,
            "order" => 12,
            "name" => "Tarifas de Vale-Transporte",
            "fileServerPath" => "/12_TarifasValeTransporte",
        ],
        [
            "id" => 13,
            "order" => 13,
            "name" => "Vale-Transporte do Vínculo",
            "fileServerPath" => "/13_ValeTransporteVinculo",
        ],
        [
            "id" => 14,
            "order" => 14,
            "name" => "Beneficiário",
            "fileServerPath" => "/14_Beneficiario",
        ],
        [
            "id" => 15,
            "order" => 15,
            "name" => "Conta Corrente",
            "fileServerPath" => "/15_ContaCorrente",
        ],
        [
            "id" => 16,
            "order" => 16,
            "name" => "Forma de Comunicação",
            "fileServerPath" => "/16_FormaComunicacao",
        ],
        [
            "id" => 17,
            "order" => 17,
            "name" => "Horários",
            "fileServerPath" => "/17_Horarios",
        ],
    ];

    public function __construct(
        EntityManager $entityManager,
        HCMParameters $hcmParameters
    ) {
        $this->entityManager = $entityManager;
        $this->hcmParameters = $hcmParameters;
    }

    public function uploadMigracaoPlanilha(array $row): ?string
    {
        $migracao = $this->getMigracao($row["NRMIGRACAO"]);
        if (empty($migracao)) {
            return null;
        }
        if ($this->migracaoPossuiArquivo($migracao)) {
            throw new \Exception(Messages::HCM00000_GPE_MIGRACAO_FILE_EXISTS);
        }
        return $this->uploadMigracaoPlanilhaFile($row, $migracao);
    }

    public function uploadMigracaoPlanilhaFile(
        array $rowWithDsPath,
        GpeMigracao $migracao
    ): ?string {
        $filePath = $this->buildPlanilhaFilePath(
            $migracao->getNrmigracao(),
            $migracao->getIdstatus()
        );
        FileUploader::uploadFilesInRow($rowWithDsPath, "DSPATH", $filePath);
        return $rowWithDsPath["DSPATH"][0]["path"] ?? null;
    }

    public function getMigracaoPlanilhaLink(string $nrMigracao): ?string
    {
        $migracao = $this->getMigracao($nrMigracao);
        $migracaoOrFileNonExistent =
            empty($migracao) || !$this->migracaoPossuiArquivo($migracao);
        if ($migracaoOrFileNonExistent) {
            return null;
        }
        return Util::getFileServerUrl() . $migracao->getDsarquivo();
    }

    public function getModeloPlanilhaLink(): ?string
    {
        $anexo = $this->getAnexoModelo();
        if (empty($anexo)) {
            return null;
        }
        return Util::getFileServerUrl() . $anexo->getDspath();
    }

    public function deleteMigracaoPlanilha(string $nrMigracao): void
    {
        $migracao = $this->getMigracao($nrMigracao);
        if (empty($migracao)) {
            return;
        }
        $this->removeMigracaoFile($migracao);
        $this->resetMigracao($migracao);
    }

    public function buildPlanilhaFilePath(
        string $nrMigracao
    ): string {
        $migracaoConfig = $this->findMigracaoConfig($nrMigracao);
        $subDirPath = $migracaoConfig["fileServerPath"] ?? "";
        return self::FILE_SERVER_BASE_PATH . $subDirPath;
    }

    public function getMigracao(string $nrMigracao): ?GpeMigracao
    {
        $findConfig = [
            "nrorg" => $this->hcmParameters->getNrorgTrabalho(),
            "nrmigracao" => $nrMigracao,
        ];
        return $this->entityManager
            ->getRepository(Repositories::GPE_MIGRACAO)
            ->findOneBy($findConfig);
    }

    public function getAnexoModelo(): ?GpeAnexos
    {
        $findConfig = [
            "nrorg" => $this->hcmParameters->getNrorgPadrao(),
            "nrparcnegocio" => Util::DEFAULT_NRPARCNEGOCIO_ORG_PADRAO,
            "nmanexo" => self::DEFAULT_MODELO_FILENAME,
        ];
        return $this->entityManager
            ->getRepository(Repositories::GPE_ANEXOS)
            ->findOneBy($findConfig);
    }

    public function setArquivoMigracao(
        string $nrMigracao,
        string $nmArquivo,
        string $dsArquivo
    ): void {
        $migracao = $this->getMigracao($nrMigracao);
        if (empty($migracao)) {
            return;
        }
        $migracao->setNmarquivo($nmArquivo);
        $migracao->setDsarquivo($dsArquivo);
        $this->entityManager->persist($migracao);
        $this->entityManager->flush($migracao);
    }

    public function resetMigracao(GpeMigracao $migracao): void
    {
        $migracao->setQtdadospendentes(0);
        $migracao->setDsarquivo(null);
        $migracao->setNmarquivo(null);
        $this->entityManager->persist($migracao);
        $this->entityManager->flush($migracao);
    }

    public function iniciarMigracaoPlanilha(): void
    {
        if ($this->hasMigracaoData()) {
            throw new \Exception(
                Messages::HCM00000_GPE_MIGRACAO_ALREADY_STARTED
            );
        }
        $this->createMigracaoData();
    }

    public function createMigracaoData(): void
    {
        foreach (self::TIPO_MIGRACAO_CONFIG as $migracaoConfig) {
            $this->setNovaMigracao($migracaoConfig);
        }
    }

    public function setNovaMigracao(array $migracaoConfig): void
    {
        $nrMigracao = $migracaoConfig["id"] ?? null;
        $nomeMigracao = $migracaoConfig["name"] ?? null;
        $ordemMigracao = $migracaoConfig["order"] ?? null;

        $gpeMigracao = Repositories::GPE_MIGRACAO;
        $novaMigracao = new $gpeMigracao();
        $novaMigracao->setNrorg($this->hcmParameters->getNrorgTrabalho());
        $novaMigracao->setNrmigracao($nrMigracao);
        $novaMigracao->setNrordem($ordemMigracao);
        $novaMigracao->setIdstatus(self::DEFAULT_MIGRACAO_IDSTATUS);
        $novaMigracao->setDsmigracao($nomeMigracao);
        $novaMigracao->setQtdadospendentes(0);
        $novaMigracao->setQtdadosimportados(0);
        $novaMigracao->setDsarquivo(null);
        $novaMigracao->setNmarquivo(null);
        $this->entityManager->persist($novaMigracao);
        $this->entityManager->flush($novaMigracao);
    }

    public function removeMigracaoFile(GpeMigracao $migracao): void
    {
        $migracaoNaoPossuiArquivoRelacionado = !$this->migracaoPossuiArquivo(
            $migracao
        );
        if ($migracaoNaoPossuiArquivoRelacionado) {
            return;
        }
        FileUploader::deleteFile($migracao->getDsarquivo());
    }

    public function migracaoPossuiArquivo(GpeMigracao $migracao): bool
    {
        return !empty($migracao->getDsarquivo());
    }

    public function hasMigracaoData(): bool
    {
        $findConfig = ["nrorg" => $this->hcmParameters->getNrorgTrabalho()];
        return !empty(
            $this->entityManager
                ->getRepository(Repositories::GPE_MIGRACAO)
                ->findBy($findConfig)
        );
    }

    public function findMigracaoConfig(string $nrMigracao): array
    {
        $migracaoConfig = array_filter(self::TIPO_MIGRACAO_CONFIG, function (
            $item
        ) use ($nrMigracao) {
            return $item["id"] == $nrMigracao;
        });
        return current($migracaoConfig) ?? [];
    }

    private $jsonPath;
    private $params;
    private $nomePlanilha;
    private $urlPlanilha;
    private $importacao;
    private $nomeImportacao;

    public function importarPlanilha($row): void
    {
        $this->prepareImportacao($row);

        File::downloadMidia($this->nomePlanilha, $this->urlPlanilha);

        $this->importacao->importar($this->nomePlanilha);

        $this->setStatusMigracao();

        $this->setQtdLinhasPendentes();

        $this->setQtdLinhasImportadas();
    }

    public function prepareImportacao($row): void
    {
        $planilhaDontExists =
            is_null($row["DSARQUIVO"]) ||
            $row["DSARQUIVO"] == "" ||
            $row["DSARQUIVO"] == "null";

        if ($planilhaDontExists) {
            throw new \Exception(
                "Nenhuma planilha foi encontrada para importação!"
            );
        }
        
        $row["DSMIGRACAO"] = Normalize::noAccents($row["DSMIGRACAO"]);
        $row["DSMIGRACAO"] = Normalize::camelCase($row["DSMIGRACAO"]);

        $this->jsonPath = Path::IMPORTACAO . $row["DSMIGRACAO"] . ".json";
        $this->nomePlanilha = explode("/", $row["DSARQUIVO"])[5];
        $this->urlPlanilha = $row["DSARQUIVO"];
        $this->nomeImportacao = $row["DSMIGRACAO"];
        $this->errorsIgnore = $row["errorsIgnore"];
        $this->params = [
            "NRORG" => $this->hcmParameters->getNrorgTrabalho(),
            "NRORG_PADRAO" => $this->hcmParameters->getNrorgPadrao(),
            "CDOPERADOR" => $this->hcmParameters->getCdoperador(),
            "NRMIGRACAO" => $row["NRMIGRACAO"],
        ];

        $this->importacao = new Importacao(
            $this->jsonPath,
            $this->params,
            $this->entityManager,
            $this->nomeImportacao,
            $this->errorsIgnore
        );
    }

    private function setStatusMigracao(): void
    {
        $findConfig = [
            "nrorg" => $this->hcmParameters->getNrorgTrabalho(),
            "nrmigracao" => $this->params["NRMIGRACAO"],
        ];

        $gpeMigracao = $this->entityManager
            ->getRepository(Repositories::GPE_MIGRACAO)
            ->findOneBy($findConfig);

        $this->entityManager->beginTransaction();

        if (!empty($this->errosImportacaoPlanilha())) {
            $gpeMigracao->setIdstatus("E");
        } else {
            $gpeMigracao->setIdstatus("C");
        }

        $this->entityManager->flush();
        $this->entityManager->commit();
    }

    //TODO: merge
    private function setQtdLinhasPendentes()
    {
        $findConfig = [
            "nrorg" => $this->hcmParameters->getNrorgTrabalho(),
            "nrmigracao" => $this->params["NRMIGRACAO"],
        ];

        $gpeMigracao = $this->entityManager
            ->getRepository(Repositories::GPE_MIGRACAO)
            ->findOneBy($findConfig);

        $erros = $this->errosImportacaoPlanilha();

        $this->entityManager->beginTransaction();

        if (!empty($erros)) {
            $gpeMigracao->setQtdadospendentes(count($erros));
        } else {
            $gpeMigracao->setQtdadospendentes(0);
        }

        $this->entityManager->flush();
        $this->entityManager->commit();
    }
    
    //TODO: merge
    private function setQtdLinhasImportadas()
    {
        $findConfig = [
            "nrorg" => $this->hcmParameters->getNrorgTrabalho(),
            "nrmigracao" => $this->params["NRMIGRACAO"],
        ];

        $gpeMigracao = $this->entityManager
            ->getRepository(Repositories::GPE_MIGRACAO)
            ->findOneBy($findConfig);

        $this->entityManager->beginTransaction();

        $gpeMigracao->setQtdadosimportados(
            $this->importacao->getQtdLinhasImportadas()
        );

        $this->entityManager->flush();
        $this->entityManager->commit();
    }

    public function errosImportacaoPlanilha(): array
    {
        $erros = $this->importacao->getErrosImportacao();

        return $erros ?? [];
    }

    public function downloadLogImportacao($logPath): string
    {
        $log = new Log($logPath);

        return $log->get() ?? "";
    }
}

class Importacao extends ImportacaoXLS
{
    public function importar($nomePlanilha): void
    {
        $this->loadPlanilha(Path::TEMP . $nomePlanilha);

        unlink(Path::TEMP . $nomePlanilha);

        $this->checkAllDataTyping();
        $this->applyFieldFormat();
        $this->checkFkValues();
        $this->insertDados();
    }
}

class File
{
    public static function downloadMidia($fileName, $fileUrl): void
    {
        $fh = fopen(Path::TEMP . $fileName, "w");
        $ch = curl_init();

        curl_setopt(
            $ch,
            CURLOPT_URL,
            "http://midia.teknisa.com/files/" . $fileUrl
        );
        curl_setopt($ch, CURLOPT_FILE, $fh);
        curl_exec($ch);
        curl_close($ch);
    }
}

class Path
{
    const TEMP = "/tmp/";
    const IMPORTACAO = "/ImportacaoXLS/";
}

class Log
{
    private $path;

    public function __construct($path)
    {
        $this->path = $path;
    }

    public function get()
    {
        $stream = fopen($this->path, "r");

        $data = stream_get_contents($stream);

        fclose($stream);

        return $data;
    }
}

class Normalize
{
    private const UTF8 = [
        "/[áàâãªä]/u" => "a",
        "/[ÁÀÂÃÄ]/u" => "A",
        "/[ÍÌÎÏ]/u" => "I",
        "/[íìîï]/u" => "i",
        "/[éèêë]/u" => "e",
        "/[ÉÈÊË]/u" => "E",
        "/[óòôõºö]/u" => "o",
        "/[ÓÒÔÕÖ]/u" => "O",
        "/[úùûü]/u" => "u",
        "/[ÚÙÛÜ]/u" => "U",
        "/ç/" => "c",
        "/Ç/" => "C",
        "/ñ/" => "n",
        "/Ñ/" => "N",
        "/–/" => "-", // UTF-8 hyphen to "normal" hyphen
        "/[’‘‹›]/u" => " ", // Literally a single quote
        "/[“”«»„]/u" => " ", // Double quote
        "/ /" => " ", // nonbreaking space (equiv. to 0x160)
    ];

    public static function noAccents($string)
    {
        return preg_replace(array_keys(self::UTF8), array_values(self::UTF8), $string);
    }

    public static function camelCase($string)
    {
        return str_replace(" ", "", ucwords(str_replace("_", " ", $string)));
    }
}